/*
  A theory of a simple JVM.
  (setq tab-width 2)
 */
THEORY Jvm

DEFINITIONS

DATATYPES
/* The basic value which we can store.
   (an integer or a reference) */
         value       = nat;

/* Identifiers for classes, methods. Includes (for methods) 
   the informations needed locally to handle argument-passing. */
datatype mId         = mId      of nat    /* Id                      */
                                 * nat;   /* NArgs                   */
datatype clsId       = clsId    of nat;   /* Id                      */

/* The basic instructions modelled. */
datatype instruction = iadd
                     | isub
                     | nop
                     | dup
                     | swap
                     | pop
                     | push     of value
                     | goto     of nat
                     | iceq     of nat
                     | iclt     of nat
                     | load     of nat
                     | store    of nat
                     | new      of clsId
                     | putfield of nat
                     | getfield of nat
                     | invoke   of mId
                     | return
                     | vreturn
                     | start    of mId
                     | monitorenter
                     | monitorexit
                     ;

/* Some mnemonic names. */
         code        = instruction list;
         pc          = nat;
         vstack      = value list;
         locals      = value list;

/* The possible constants (methods, classes) */
         methdef     = code               /* code for method.        */
                     * nat                /* #locals                 */
                     * pc                 /* start value of PC       */
                     ;
datatype method      = method of nat      /* Id                      */
                               * methdef  /* Method definition.      */
                     ;
datatype class       = class  of clsId    /* Id                      */
                               * nat      /* #fields                 */
                     ;
         constpool   = (class list)
                     * (method list)
                     ;

/* A frame for a method invocation instance. */
datatype frame       = frame of code   *  /* code for method.        */
                                pc     *  /* current program counter */
                                vstack *  /* stack of values         */
                                locals;   /* local variables         */

/* The actions that may appear on arrows. */
datatype action      = tau
                     | call     of mId    /* Method called.          */
                                 * vstack /* Arguments passed.       */
                     | result
                     | vresult  of value  /* Result returned.        */
                     | allocate of clsId  /* What to allocate.       */
                                 * nat    /* ref to new              */
                     | get      of nat    /* ref from                */
                                 * nat    /* field index             */
                                 * value  /* put into                */
                     | put      of nat    /* ref from                */
                                 * nat    /* field index             */
                                 * value  /* the value               */
                     | spawn    of nat    /* ref to started instance */
                                 * mId    /* method to start in      */
                     | enter    of nat    /* ref to class with lock  */
                     | exit     of nat    /* ref to class with lock  */
                     ;

/* A lock (in effect, just a boolean with a nicer name). */
datatype lock        = locked
                     | unlocked
                     ;


/* Parts of a system. */
         framel      = frame list;
datatype thread      = t        of nat    /* Reference to the class  */
                                 * framel /* The list of frames.     */
                     ;
         threadl     = thread list;
         classInst   =             clsId   /* The Id of the class.   */
                                 * locals  /* Locals of the object.  */
                                 * lock    /* The associated mutex.  */
                                 * threadl /* The threads waiting    */
                     ;                     /* for the lock.          */
         heap        = classInst list;

/* The type of a complete system. */
datatype system      = system   of threadl /* The currently active   */
                                           /* threads.               */
                                 * heap    /* The heap of objects.   */
                                 * constpool/*Available contants.    */
END /* DATATYPES  */


/* Accessor functions. Needed in array.inc */
SML FROM "JvmAccess.sml"


/* Predicates describing lists as arrays,
   along with the associated tactics */
#include <array.inc>
SML FROM "array_tactics.sml"


PREDICATES
/* The method transistion relation. */
LET
  methodTrans : frame --> action --> frame --> prop =
    \F1 : frame . \A : action . \F2 : frame .
      cases F1 of
           frame(CS, PC, VS, LS) =>
           exists Inst : instruction .
                 (at_i CS PC Inst) and
                 cases Inst of
                   push(N)        => pushTrans     Inst F1 A F2
                 | pop()          => popTrans      Inst F1 A F2
                 | dup()          => dupTrans      Inst F1 A F2
                 | swap()         => swapTrans     Inst F1 A F2
                 | nop()          => nopTrans      Inst F1 A F2
                 | iadd()         => iaddTrans     Inst F1 A F2
                 | isub()         => isubTrans     Inst F1 A F2
                 | goto(N)        => gotoTrans     Inst F1 A F2
                 | iceq(N)        => iceqTrans     Inst F1 A F2
                 | iclt(N)        => icltTrans     Inst F1 A F2
                 | store(N)       => storeTrans    Inst F1 A F2
                 | load(N)        => loadTrans     Inst F1 A F2
                 | invoke(MId)    => invokeTrans   Inst F1 A F2
                 | return()       => returnTrans   Inst F1 A F2
                 | vreturn()      => vreturnTrans  Inst F1 A F2
                 | new(ClsId)     => newTrans      Inst F1 A F2
                 | putfield(N)    => putfieldTrans Inst F1 A F2
                 | getfield(N)    => getfieldTrans Inst F1 A F2
                 | start(MId)     => startTrans    Inst F1 A F2
                 | monitorexit()  => mexitTrans    Inst F1 A F2
                 | monitorenter() => menterTrans   Inst F1 A F2
                 end /* cases Inst */
      end /* cases F1 */
WHERE
  pushTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .
      cases (Inst, F1, A, F2) of
        (push(Arg),
         frame(CS, PC, VS, LS),
         tau(),
         frame(CS, PC+1, [Arg | VS], LS)
        )
      end /* cases */
  END; /* pushTrans */
  
  popTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (pop(),
         frame(CS, PC, [N | VS], LS),
         tau(),
         frame(CS, PC+1, VS, LS)
        )
      end /* cases */
  END; /* popTrans */
  
  dupTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (dup(),
         frame(CS, PC, [N | VS], LS),
         tau(),
         frame(CS, PC+1, [N, N | VS], LS)
        )
      end /* cases */
  END; /* dupTrans */
  
  swapTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (swap(),
         frame(CS, PC, [N1, N2 | VS], LS),
         tau(),
         frame(CS, PC+1, [N2, N1 | VS], LS)
        )
      end /* cases */
  END; /* swapTrans */
  
  nopTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (nop(),
         frame(CS, PC, VS, LS),
         tau(),
         frame(CS, PC+1, VS, LS)
        )
      end /* cases */
  END; /* nopTrans */
  
  iaddTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (iadd(),
         frame(CS, PC, [N1, N2 | VS], LS),
         tau(),
         frame(CS, PC+1, [N | VS], LS)
        ) =>
          natPlus N1 N2 N
      end /* cases */
  END; /* iaddTrans */

  isubTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (isub(),
         frame(CS, PC, [N1, N2 | VS], LS),
         tau(),
         frame(CS, PC+1, [N | VS], LS)
        ) =>
          natMinus N1 N2 N
      end /* cases */
  END; /* isubTrans */
    
  gotoTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (goto(Arg),
         frame(CS, PC, VS, LS),
         tau(),
         frame(CS, Arg, VS, LS)
        )
      end /* cases */
  END; /* gotoTrans */
    
  iceqTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A) of
        (iceq(Arg), frame(CS, PC, [N1, N2 | VS], LS), tau())
        =>
          if (N1 = N2)
          then (
               F2 = frame(CS, Arg, VS, LS)
          ) else (
               F2 = frame(CS, PC+1, VS, LS)
          )
      end /* cases */
  END; /* iceqTrans */

  icltTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A) of
        (iclt(Arg), frame(CS, PC, [N1, N2 | VS], LS), tau())
        =>
          if (natLess N1 N2)
          then (
               F2 = frame(CS, Arg, VS, LS)
          ) else (
               F2 = frame(CS, PC+1, VS, LS)
          )
      end /* cases */
  END; /* icltTrans */

  storeTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (store(Arg), 
         frame(CS, PC, [Val | VS], LS), 
         tau(),
         frame(CS, PC+1, VS, LS'))
        =>
            set_n LS Arg Val LS'
      end /* cases */
  END; /* storeTrans */

  loadTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (load(Arg), 
         frame(CS, PC, VS, LS), 
         tau(),
         frame(CS, PC+1, [Val | VS], LS))
        =>
            at_n LS Arg Val
      end /* cases */
  END; /* loadTrans */

  invokeTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (invoke(mId(Id, NArgs)), 
         frame(CS, PC, VS, LS), 
         call(mId(Id, NArgs), Args),
         frame(CS, PC+1, VS', LS))
        => 
           split_rev_n VS NArgs Args VS'
      end /* cases */
  END; /* invokeTrans */

  returnTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (return(), 
         frame(CS, PC, VS, LS), 
         result(),
         frame(CS, PC+1, VS', LS))
      end /* cases */
  END; /* returnTrans */

  vreturnTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (vreturn(), 
         frame(CS, PC, [V |VS], LS), 
         vresult(V),
         frame(CS, PC+1, VS, LS))
      end /* cases */
  END; /* vreturnTrans */

  getfieldTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (getfield(Idx), 
         frame(CS, PC, [Ref | VS], LS), 
         get(Ref, Idx, Val),
         frame(CS, PC+1, [Val | VS], LS))
      end /* cases */
  END; /* getfieldTrans */

  putfieldTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (putfield(Idx),
         frame(CS, PC, [Val, Ref | VS], LS), 
         put(Ref, Idx, Val),
         frame(CS, PC+1, VS, LS))
      end /* cases */
  END; /* putfieldTrans */

  newTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (new(ClsId), 
         frame(CS, PC, VS, LS), 
         allocate(ClsId, Ref),
         frame(CS, PC+1, [Ref | VS], LS))
      end /* cases */
  END; /* newTrans */

  startTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (start(MId), 
         frame(CS, PC, [Ref | VS], LS), 
         spawn(Ref, MId),
         frame(CS, PC+1, VS, LS))
      end /* cases */
  END; /* startTrans */

  menterTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (monitorenter(), 
         frame(CS, PC, [Ref | VS], LS), 
         enter(Ref),
         frame(CS, PC+1, VS, LS))
      end /* cases */
  END; /* monitorenterTrans */

  mexitTrans : instruction --> frame --> action --> frame --> prop =
    \Inst : instruction . \F1 : frame . \A : action . \F2 : frame .  
      cases (Inst, F1, A, F2) of
        (monitorexit(), 
         frame(CS, PC, [Ref | VS], LS), 
         exit(Ref),
         frame(CS, PC+1, VS, LS))
      end /* cases */
  END /* monitorexitTrans */

END; /* LET, methodTrans */

/* The system transition relation. */
LET
transRel : system --> action --> system --> prop  =
  \S1 : system . \A : action . \S2 : system .
  cases (S1, A) of
  (system(Ths, H, CP), tau()) =>
    exists NThreads : nat .
      size_t Ths NThreads+1       and
      transRelImpl S1 S2 NThreads
  end /* cases */
WHERE
  LET
  transRelImpl : system --> system --> nat --> prop <=
    \S1 : system . \S2 : system . \NThread : nat .
      cases S1 of
      system(Ths, H, CP) =>
          (exists Ths' : thread list .
           exists H'   : heap        .
             transOneThread NThread Ths H CP Ths' H' and
             S2 = system(Ths', H', CP)
          )
          \/
          (cases NThread of
             0          => ff
           | NThread'+1 => transRelImpl S1 S2 NThread'
           end /* cases */
          )
      end /* cases, S1 */
  WHERE
    transOneThread : nat     --> thread list --> heap --> 
                   constpool --> thread list --> heap --> prop  =
      \NThread : nat  . \Ths  : thread list . \H1 : heap . 
      \CP : constpool . \Ths' : thread list . \H2 : heap .
      exists T1 : thread .
      exists T2 : thread .
        at_t Ths NThread T1 and
        cases T1 of
          t(TId, [F | Tail]) =>
          exists A' : action . 
          exists F' : frame  .
            methodTrans F A' F' and
            cases A' of
              tau()                => 
              H2 = H1 and
              set_t Ths NThread t(TId, [F' | Tail]) Ths'
            | result()             =>
              H2 = H1 and 
              if Tail = []:frame list then
                remove_t Ths NThread Ths'
              else
                set_t Ths NThread t(TId, Tail) Ths'
            | vresult(N)           =>
              cases Tail of
              [frame(CS, PC, VS, LS) | Tail'] =>
                T2 = t(TId, 
                       [frame(CS, PC, [N | VS], LS) | Tail']) and
                H2 = H1                                       and 
                set_t Ths NThread T2 Ths'
              end /* cases Tail */
            | call(MId, Args)      =>
              exists Code    : code .
              exists Nlocals : nat  .
              exists StartPC : nat  .
              cases (MId, CP) of (mId(Id, NArgs), (C, M)) =>
                contains_me M Id (Code, Nlocals, StartPC) and
                (exists LS  : locals . 
                 exists LS' : locals .
                  create_n Nlocals 0 LS    and
                  set_from_n LS 0 Args LS' and
                  T2 = t(TId, [frame(Code, StartPC, []:vstack, LS'), 
                        F' | Tail])        and
                  H2 = H1                  and
                  set_t Ths NThread T2 Ths')
              end
            | get(Ref, Idx, Val)   =>
              exists ClsId  : clsId       .
              exists Fields : locals      .
              exists L      : lock        .
              exists Q      : thread list .
                at_h H1 Ref (ClsId, Fields, L, Q) and
                at_n Fields Idx Val               and
                H2 = H1                           and
                set_t Ths NThread t(TId, [F' | Tail]) Ths'
            | put(Ref, Idx, Val)   =>
              exists Fields  : locals .
              exists Fields' : locals .
              exists ClsId   : clsId  .
              exists L      : lock        .
              exists Q      : thread list .
                at_h H1 Ref (ClsId, Fields, L, Q)      and
                set_n Fields Idx Val Fields'           and
                set_h H1 Ref (ClsId, Fields', L, Q) H2 and
                set_t Ths NThread t(TId, [F' | Tail]) Ths'
            | allocate(ClsId, Ref) =>
              exists Fields  : locals .
              exists NFields : nat    .
              cases CP of (C, M) =>
                contains_cl C ClsId NFields                 and
                create_n  NFields 0 Fields                  and
                extend_h H1 (ClsId,      Fields, 
                             unlocked(), []:thread list) H2 and
                size_h H2 Ref+1                             and
                set_t Ths NThread t(TId, [F' | Tail]) Ths'
              end
            | spawn(Ref, RunId)    =>
              exists Code    : code        .
              exists Nlocals : nat         .
              exists StartPC : nat         .
              exists LS      : locals      .
              exists LS'     : locals      .
              exists Thszz   : thread list .
              cases (RunId, CP) of (mId(Id, NArgs),(C, M)) =>
                contains_me M Id (Code, Nlocals, StartPC)      and
                create_n Nlocals 0 LS                          and
                set_n LS 0 Ref LS'                             and
                T2 = t(Ref, 
                       [frame(Code, StartPC, []:vstack, LS')]) and
                extend_t Ths T2 Thszz                          and
                H2 = H1                                        and
                set_t Thszz NThread t(TId, [F' | Tail]) Ths'
              end
            | enter(Ref)            =>
              exists A : clsId       .
              exists B : locals      .
              exists L : lock        .
              exists Q : thread list .
                at_h H1 Ref (A, B, L, Q) and
                if L = locked() then (
                  exists Q' : thread list .
                    extend_t Q t(TId, [F' | Tail]) Q'     and
                    set_h H1 Ref (A, B, L, Q') H2 and
                    remove_t Ths NThread Ths' 
                ) else (
                  set_h H1 Ref (A, B, locked(), Q) H2 and
                  set_t Ths NThread t(TId, [F' | Tail]) Ths'
                )
            | exit(Ref)             =>
              exists A : clsId       .
              exists B : locals      .
              exists Q : thread list .
                /* NOTE: is this true? i.e. do we really have */
                at_h H1 Ref (A, B, locked(), Q) and
                cases Q of
                  []:thread list =>
                    set_h H1 Ref (A, B, unlocked(), Q) H2 and
                    set_t Ths NThread t(TId, [F' | Tail]) Ths'
                | [T | QTail]    =>
                  exists Thszz : thread list .
                    set_h H1 Ref (A, B, locked(), QTail) H2 and
                    extend_t Ths T Thszz                    and
                    set_t Thszz NThread t(TId, [F' | Tail]) Ths'
                end /* cases Q */
            end /* cases A' */
        end /* cases T1 */
    end /* transOneThread */
  END /* LET, transRelImpl */
END /* LET, transRel */

END /* PREDICATES */

MODALITIES
  transRel
END /* MODALITIES */


/* Usefull abbreviations to help formulaconstruction.
   (always(phi), ...) */
#include <patterns.pde>

/* Tactics for general simplification. */
SML FROM "simplify.sml"

/* Tactics for specific simplification of the transitionrelation. */
SML FROM "tactics.sml"


#include <fairex>

/*
#include <apprenticeex>
*/

END /* THEORY     */
